#!/usr/bin/env python

"""
A handler to help with testing.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from imiptools.client import ClientForObject
from imiptools.data import Object, get_address, parse_object
from imiptools.dates import get_datetime, to_timezone
from imiptools.mail import Messenger
from imiptools.period import RecurringPeriod
import imip_store
import sys

class TestClient(ClientForObject):

    """
    A content handler for use in testing, as opposed to operating within the
    mail processing pipeline.
    """

    # Action methods.

    def handle_request(self, action, start=None, end=None):

        """
        Process the current request for the current user. Return whether the
        given 'action' was taken.

        If 'start' and 'end' are specified, they will be used in any
        counter-proposal.
        """

        # Reply only on behalf of this user.

        if action in ("accept", "decline"):
            attendee_attr = self.update_participation(action == "accept" and "ACCEPTED" or "DECLINED")
            method = "REPLY"

        # For counter-proposals, set a new main period for the event.

        elif action == "counter":
            attendee_attr = self.obj.get_value_map("ATTENDEE").get(self.user)
            period = self.obj.get_main_period(self.get_tzid())

            # Use the existing or configured time zone for the specified
            # datetimes.

            start = to_timezone(get_datetime(start), period.tzid)
            end = to_timezone(get_datetime(end), period.tzid)
            period = RecurringPeriod(start, end, period.tzid, period.origin, period.get_start_attr(), period.get_end_attr())
            self.obj.set_period(period)
            method = "COUNTER"
        else:
            return None

        if not attendee_attr:
            return None

        # NOTE: This is a simpler form of the code in imipweb.client.

        organiser = get_address(self.obj.get_value("ORGANIZER"))

        self.update_sender(attendee_attr)
        self.obj["ATTENDEE"] = [(self.user, attendee_attr)]
        self.update_dtstamp()
        self.update_sequence(False)

        message = self.messenger.make_outgoing_message(
            [self.obj.to_part(method)],
            [organiser],
            outgoing_bcc=get_address(self.user)
            )

        return message.as_string()

# A simple main program that attempts to handle a stored request, writing the
# response message to standard output.

if __name__ == "__main__":
    try:
        action, store_dir, preferences_dir, user = sys.argv[1:5]
        if action == "counter":
            start, end = sys.argv[5:7]
            i = 7
        else:
            start, end = None, None
            i = 5
        uid, recurrenceid = (sys.argv[i:i+2] + [None] * 2)[:2]
    except ValueError:
        print >>sys.stderr, """\
Need 'accept', 'counter' or 'decline', a store directory, a preferences
directory, user URI, any counter-proposal datetimes (see below), plus the
appropriate event UID and RECURRENCE-ID (if a recurrence is involved).

The RECURRENCE-ID must be in exactly the form employed by the store, not a
different but equivalent representation.

Alternatively, omit the UID and RECURRENCE-ID and provide event-only details on
standard input to force the script to handle an event not already present in the
store.

If 'counter' has been indicated, alternative start and end datetimes are also
required.
"""
        sys.exit(1)

    store = imip_store.FileStore(store_dir)

    if uid is not None:
        fragment = store.get_event(user, uid, recurrenceid)

        if not fragment:
            print >>sys.stderr, "No such event:", uid, recurrenceid
            sys.exit(1)
    else:
        fragment = parse_object(sys.stdin, "utf-8")

    obj = Object(fragment)
    handler = TestClient(obj, user, Messenger(), store, None, preferences_dir)
    response = handler.handle_request(action, start, end)

    if response:
        if uid is not None:
            store.dequeue_request(user, uid, recurrenceid)
        print response
    else:
        sys.exit(1)

# vim: tabstop=4 expandtab shiftwidth=4
